﻿/* Documentation
 * 
 * Developer: Thomas Wilson (TappaTekie)
 * Signed: 28/03/13 04:08 GMT+0
 *
 * Description
 * ===========
 * Serializes and deserializes objects to and from a byte representation
 * 
 * Notes
 * =======
 *  Supports
 *  ========
 *   1) Value types: sbyte, byte, short, ushort, int, uint, long, ulong
 *                 string, char, bool, double, float and decimal
 *   2) Enums
 *   3) 1 dimensional arrays (does not allow jagged arrays)
 *   4) Date and Time
 *   5) CLR object with properties and types with the supported serializable types
 *          (Uses reflection)
 *   6) Additional serialization systems can be added
 *   7) Shallow and deep object cloning using serialization
 * 
 *  For now, this class can only cope with primitive objects to serialize but is
 *  wide open to addons.
 *  
 *  Caution: Serializing a float, decimal or double that is larger than a 32bit 
 *  value leads to value type overflow issues
 *  
 *  If a type cannot be recognized, the Null type is cast to it which contains
 *  no information about the original value, it is only there as a placeholder to
 *  specify that a property or field was present in object serialization.
 *  
 *  SerializeLength is the core function which writes an integer to a stream by
 *  converting it into binary form and splitting it into 8bit chunks which then
 *  get written to the stream to then be rebuilt by DeSerializeLength. These functions
 *  ignore the SerializeType
 *  
 *  A security notice would be the use of RegisterSerializer by third party applications
 *  as well as using objects to store secured information which can then be intercepted by
 *  the registered serializer class externally.
 *  A workaround to this is to ONLY use SerializeObject to serialize and deserialize secured classes 
 *  which override the registered serializers completely.
*/

using System;
using System.IO;
using System.Text;
using System.Reflection;

namespace CodeSample {
    public static partial class Serializer {
        private static ISerializer[] p_Serializers = new ISerializer[0];
        private static object p_SerializerSync = new object();
        private static string p_Token = "SOBJEF";
        private static string p_Version = "1.0";

        public static void WriteFileToken(Stream stream) { 
            //write the file token to identify that a file is dedicated storing serialized objects
            string token = p_Token + "_V" + p_Version;
            stream.Write(Encoding.ASCII.GetBytes(token), 0, token.Length);
        }
        public static void SkipFileToken(Stream stream) { 
            //position must be 0
            if (stream.Position != 0) { return; }

            //check if the stream actually has a file token
            if (!IsSerializedStream(stream)) { return; }

            //skip over the token
            stream.Position = p_Token.Length + p_Version.Length + 2;
        }
        public static bool IsSerializedStream(Stream stream) { 
            /*
                This function does not check the file version
                When these functions are changed in the way that will effect 
                previously formatted serialized files, this function will need to 
                check it.
             */

            //buffer the stream position so we can read the beginning of the stream
            long positionBuffer = stream.Position;
            stream.Seek(0, SeekOrigin.Begin);

            //get what the token should be
            string token = p_Token + "_V" + p_Version;

            //can the stream contain the token?
            if (stream.Length < token.Length) {
                stream.Position = positionBuffer;
                return false; 
            }

            //read the token from the stream
            byte[] compare = new byte[token.Length];
            stream.Read(compare, 0, compare.Length);

            //seek back to the buffer position and compare the tokens
            stream.Position = positionBuffer;
            return Helpers.CompareBytes(
                    compare,
                    Encoding.ASCII.GetBytes(token));
        }

        public static bool RegisterSerializer(ISerializer serializer) {
            //do not accept null references
            if (serializer == null) { return false; }

            //is the type defined for the serializer a generic type?
            if (Enum.GetName(typeof(SerializeType), serializer.TypeToken) != null) {
                return false;
            }

            lock (p_SerializerSync) {
                string name = serializer.Name.ToLower();
                byte typeToken = serializer.TypeToken;

                //does the serializer already exist?
                for (int c = 0; c < p_Serializers.Length; c++) {
                    if (typeToken == p_Serializers[c].TypeToken ||
                        name == p_Serializers[c].Name.ToLower()) {
                            return false;
                    }
                }

                //add it
                Array.Resize(ref p_Serializers, p_Serializers.Length + 1);
                p_Serializers[p_Serializers.Length - 1] = serializer;
            }

            //it was successfully added
            return true;
        }
        
        public static object DeepClone(object obj) {
            return Clone(obj, true);
        }
        public static object ShallowClone(object obj) {
            return Clone(obj, false);
        }
        public static object Clone(object obj, bool deepClone) {
            //null?
            if (obj == null) { return null; }

            //serialize the object 
            Stream ms = new MemoryStream();
            Serialize(obj, ms, deepClone);

            //deserialize from the stream to create a clone of the object
            ms.Position = 0;
            object buffer = DeSerialize(ms, obj.GetType());

            //clean up
            ms.Dispose();
            return buffer;
        }

        #region Base functions
        public static void SerializeData(SerializeType type, byte[] data, Stream stream) { 
            //write the type 
            stream.WriteByte((byte)type);

            //if the type is null, do not write anything else
            if (type == SerializeType.Null) { return; }

            //write the length of the data
            if (requireLength(type)) {
                SerializeLength((ulong)data.Length, stream);
            }

            //write the data
            stream.Write(data, 0, data.Length);
        }
        public static byte[] DeSerializeData(Stream stream, out SerializeType type) { 
            //read the type
            type = (SerializeType)stream.ReadByte();

            //null?
            if (type == SerializeType.Null) { return new byte[0]; }

            //length required?
            ulong length = 1;
            if (requireLength(type)) {
                length = DeSerializeLength(stream);
            }

            //read the bytes
            byte[] data = new byte[length];
            stream.Read(data, 0, data.Length);
            return data;
        }

        public static SerializeType GetNextType(Stream stream) { 
            //read the byte from the stream and seek back
            byte r = (byte)stream.ReadByte();
            stream.Seek(-1, SeekOrigin.Current);

            //return the type
            return (SerializeType)r;

        }

        public static void SerializeLength(ulong length, Stream stream) { 
            //define the length as a binary form
            string binary = Helpers.ChangeNumericBase(10, 2, length.ToString());

            //make it 8bit divisable so we can seperate it into byte chunks
            while (binary.Length % 8 != 0) {
                binary = "0" + binary;
            }

            //convert each 8bit chunk into a byte array
            byte[] chunks = new byte[binary.Length / 8];
            for (int c = 0; c < chunks.Length; c++) {
                chunks[c] = Convert.ToByte(
                    Helpers.ChangeNumericBase(
                        2,
                        10,
                        binary.Substring(c * 8, 8)));
            }
            
            //write the data to the stream (length then binary data)
            stream.WriteByte((byte)chunks.Length);
            stream.Write(chunks, 0, chunks.Length);
        }
        public static ulong DeSerializeLength(Stream stream) { 
            //read the length of the raw bytes that contain the information
            int length = (int)stream.ReadByte();

            //read the raw data
            byte[] raw = new byte[length];
            stream.Read(raw, 0, raw.Length);

            //convert to a binary string 
            string binary = "";
            for (int c = 0; c < length; c++) {
                string bin = Helpers.ChangeNumericBase(10, 2, raw[c].ToString());
                while (bin.Length < 8) { bin = "0" + bin; }
                binary += bin;
            }

            //convert the binary to an integer which is now the length
            return Convert.ToUInt64(Helpers.ChangeNumericBase(2, 10, binary));
        }

        private static bool requireLength(SerializeType type) {
            switch (type) {
                case SerializeType.Boolean: return false;
            }

            return true;
        }

        public static void Serialize(object value, Stream stream) {
            Serialize(value, stream, true);
        }
        public static void Serialize(object value, Stream stream, bool deep) {
            //serialize to the appropriate type
            if (value is bool) { SerializeBoolean((bool)value, stream); return; }
            else if (value is string) { SerializeString((string)value, stream); return; }
            else if (value is char) { SerializeChar((char)value, stream); return; }
            else if (value is DateTime) { SerializeDateTime((DateTime)value, stream); return; }
            else if (value is Array) { SerializeArray((Array)value, stream, deep); return; }
            else if (value is Enum) { SerializeEnum(value, stream); return; }

            //numerical?
            if (getIntSerializerType(value) != SerializeType.Null) {
                SerializeNumeric(value, stream);
                return;
            }

            //look for a serializer that can serialize this object
            lock (p_SerializerSync) {
                for (int c = 0; c < p_Serializers.Length; c++) {
                    if (p_Serializers[c].Accepts(value, deep)) {
                        //write the type token
                        stream.WriteByte(p_Serializers[c].TypeToken);

                        //serialize it using the serializer
                        p_Serializers[c].Serialize(value, stream, deep);
                        return;
                    }
                }
            }

            //serialize as an object
            SerializeObject(value, stream, deep);
        }

        public static object DeSerialize(Stream stream) {
            return DeSerialize(stream, null);
        }
        public static object DeSerialize(Stream stream, Type objType) { 
            //get the type
            SerializeType type = GetNextType(stream);

            //integer?
            if ((byte)type > 0 && (byte)type < 20) {
                object buffer = createNumeric(type);
                DeSerializeNumeric(ref buffer, stream);
                return buffer;
            }

            //deserialize the appropriate type
            switch (type) {
                case SerializeType.Boolean: return DeSerializeBoolean(stream);
                case SerializeType.String: return DeSerializeString(stream);
                case SerializeType.Char: return DeSerializeChar(stream);
                case SerializeType.DateTime: return DeSerializeDateTime(stream);
                case SerializeType.Array: return DeSerializeArray(stream, objType);
                case SerializeType.Enum: return DeSerializeEnum(stream, objType);
                case SerializeType.Object: return DeSerializeObject(stream, objType);
            }

            //is the type unknown?
            if (Enum.GetName(typeof(SerializeType), type) == null) {
                //look for a serializer
                lock (p_SerializerSync) {
                    for (int c = 0; c < p_Serializers.Length; c++) {
                        if (p_Serializers[c].TypeToken == (byte)type) {
                            return p_Serializers[c].DeSerialize(stream, objType);
                        }
                    }
                }

            }

            //no serialization context can be found for the type
            return null;
        }

        public static void SerializeObjects(object[] values, Stream stream) {
            SerializeObjects(values, stream, true);
        }
        public static void SerializeObjects(object[] values, Stream stream, bool deep) {
            //serialize all the objects
            for (int c = 0; c < values.Length; c++) {
                Serialize(values[c], stream, deep);
            }
        }

        public static object[] DeSerializeObjects(Stream stream) { 
            //define the return buffer
            object[] buffer = new object[0];

            //deserialize all the objects
            while (stream.Position < stream.Length) {
                Array.Resize(ref buffer, buffer.Length + 1);
                buffer[buffer.Length - 1] = DeSerialize(stream);
            }

            //return the objects
            return buffer;
        }

        private static bool isGeneric(object obj) { 
            //numerical?
            if (getIntSerializerType(obj) != SerializeType.Null) { return true; }

            //look for supported types
            if (obj is bool) { return true; }
            else if (obj is char) { return true; }
            else if (obj is string) { return true; }
            else if (obj is DateTime) { return true; }
            else if (obj is Array) { return true; }
            else if (obj is Enum) { return true; }

            //not generically supported
            return false;
        }
        #endregion
        
        #region String
        public static void SerializeString(string value, Stream stream) {
            SerializeData(SerializeType.String,
                          Encoding.ASCII.GetBytes(value),
                          stream);
        }
        public static string DeSerializeString(Stream stream) { 
            //read the data
            SerializeType type;
            byte[] data = DeSerializeData(stream, out type);

            //check the type
            if (type != SerializeType.String) { return null; }

            //return the string
            return Encoding.ASCII.GetString(data);
        }
        #endregion

        #region Char
        public static void SerializeChar(char value, Stream stream) {
            SerializeData(SerializeType.Char,
                          new byte[] { (byte)value },
                          stream);
        }
        public static char DeSerializeChar(Stream stream) { 
            //read the data
            SerializeType type;
            byte[] data = DeSerializeData(stream, out type);

            //check type
            if (type != SerializeType.Char) { return (char)0; }

            //return the char
            return (char)data[0];
        }
        #endregion

        #region Boolean
        public static void SerializeBoolean(bool value, Stream stream) {
            SerializeData(SerializeType.Boolean,
                          new byte[] { (byte)(value ? 255 : 0) },
                          stream);
        }
        public static bool DeSerializeBoolean(Stream stream) { 
            //read the data
            SerializeType type;
            byte[] data = DeSerializeData(stream, out type);

            //check the type
            if (type != SerializeType.Boolean) { return false; }

            //return the boolean
            return data[0] == 255;
        }
        #endregion

        #region DateTime
        public static void SerializeDateTime(DateTime value, Stream stream) { 
            //write type
            stream.WriteByte((byte)SerializeType.DateTime);

            //write the value ticks as an integer
            SerializeLength((ulong)value.Ticks, stream);
        }
        public static DateTime DeSerializeDateTime(Stream stream) { 
            //read the type
            SerializeType type = (SerializeType)stream.ReadByte();

            //check
            if (type != SerializeType.DateTime) {
                return DateTime.MinValue;
            }

            //read the DateTime object by reading the ticks integer
            return new DateTime(
                (long)DeSerializeLength(stream));
        }
        #endregion

        #region Array
        public static void SerializeArray(Array array, Stream stream) {
            SerializeArray(array, stream, true);
        }
        public static void SerializeArray(Array array, Stream stream, bool deep) { 
            //the array must be one dimensional otherwise it won't be supported
            if (array.Rank != 1) {
                stream.WriteByte((byte)SerializeType.Null);
                return;
            }

            //convert to an object array
            object[] buffer = new object[array.Length];
            for (int c = 0; c < array.Length; c++) {
                buffer[c] = array.GetValue(c);
            }
            
            //serialize
            SerializeArray(buffer, stream, deep);
        }
        public static void SerializeArray<T>(T[] array, Stream stream) {
            SerializeArray(array, stream, true);
        }
        public static void SerializeArray<T>(T[] array, Stream stream, bool deep) { 
            //define the data to write
            Stream data = new MemoryStream();

            //iterate over every entry
            for (int c = 0; c < array.Length; c++) { 
                //serialize the object in the array
                Serialize(array[c], data, deep);
            }

            //write the header informatin
            stream.WriteByte((byte)SerializeType.Array);
            SerializeLength((ulong)array.Length, stream);

            //write the data
            byte[] raw = new byte[data.Length];
            data.Position = 0;
            data.Read(raw, 0, raw.Length);
            SerializeLength((ulong)data.Length, stream);
            stream.Write(raw, 0, raw.Length);
        }

        public static Array DeSerializeArray(Stream stream) {
            return DeSerializeArray(stream, typeof(object));
        }
        public static Array DeSerializeArray(Stream stream, Type arrayType) {
            //if the type is not specified, make it an object array
            if (arrayType == null) { arrayType = typeof(object); }

            if (arrayType.IsArray) {
                //the array type must be one dimensional           
                if (arrayType.GetArrayRank() != 1) { return null; }

                //set the array type to the underlying array object type
                arrayType = arrayType.Assembly.GetType(arrayType.FullName.Replace("[]", ""));
            }

            //read the type
            SerializeType type = (SerializeType)stream.ReadByte();
            if (type != SerializeType.Array) { return null; }

            //read the array length
            ulong arrayLength = DeSerializeLength(stream);

            //read the data length
            ulong dataLength = DeSerializeLength(stream);

            //define the return buffer and read the data
            Array buffer = Array.CreateInstance(arrayType, (int)arrayLength);
            long dataEnd = (long)(stream.Position + (long)dataLength);
            int location = 0;
            while (stream.Position != dataEnd) { 
                //de-serialize the entity
                buffer.SetValue(DeSerialize(stream), location);
                location++;
            }

            //return the array
            return buffer;            
        }
        #endregion

        #region Object
        public static void SerializeObject(object value, Stream stream) {
            SerializeObject(value, stream, true);
        }
        public static void SerializeObject(object value, Stream stream, bool deep) { 
            //define the data stream that will contain the object data
            Stream data = new MemoryStream();

            //null?
            if (value == null) { 
                //write a null type
                stream.WriteByte((byte)SerializeType.Null);
                return;
            }

            //get the public properties and fields for the object
            Type type = value.GetType();
            FieldInfo[] fields = type.GetFields();
            PropertyInfo[] properties = type.GetProperties();

            //write the fields and properties
            for (int c = 0; c < fields.Length; c++) {
                //get the value
                object v = fields[c].GetValue(value);

                //if it's not generic, if it's a deep serialization then serialize the value, otherwise write as null
                if (isGeneric(v) || deep) { Serialize(v, data); continue; }
                data.WriteByte((byte)SerializeType.Null);
            }
            for (int c = 0; c < properties.Length; c++) { 
                //property must be readable and writable
                if (!properties[c].CanWrite || !properties[c].CanRead) { continue; }

                //get the value
                object v = properties[c].GetValue(value, null);

                //if it's not generic, if it's a deep serialization then serialize the value, otherwise write as null
                if (isGeneric(v) || deep) { Serialize(v, data); continue; }
                data.WriteByte((byte)SerializeType.Null);
            }

            //write the header
            stream.WriteByte((byte)SerializeType.Object);
            SerializeLength((ulong)data.Length, stream);

            //write the data
            data.Position = 0;
            byte[] dataBytes = new byte[data.Length];
            data.Read(dataBytes, 0, dataBytes.Length);
            stream.Write(dataBytes, 0, dataBytes.Length);
        }
        
        public static object DeSerializeObject(Stream stream, Type type) { 
            //if the type is null, specify an object type
            if (type == null) { type = typeof(object); }
         
            //read the data
            SerializeType sType;
            byte[] data = DeSerializeData(stream, out sType);

            //check the type
            if (sType != SerializeType.Object) { return null; }

            //the object type must contain one constructor with no arguments
            ConstructorInfo[] typeConstructors = type.GetConstructors();
            bool valid = false;
            for (int c = 0; c < typeConstructors.Length; c++) {
                if (typeConstructors[c].GetParameters().Length == 0) {
                    valid = true;
                    break;
                }
            }
            if (!valid) { return null; }

            //convert the data read into a readable stream
            Stream dataStream = new MemoryStream();
            dataStream.Write(data, 0, data.Length);
            dataStream.Position = 0;

            //define the return object which creates a new instance of the specified type
            object buffer = Activator.CreateInstance(type);
           
            //get the field and properties
            FieldInfo[] fields = type.GetFields();
            PropertyInfo[] properties = type.GetProperties();

            //read the fields
            for (int c = 0; c < fields.Length; c++) {
                fields[c].SetValue(buffer, DeSerialize(dataStream, fields[c].FieldType));
            }

            //read the properties
            for (int c = 0; c < properties.Length; c++) { 
                //the property must be readable and writable
                if (!properties[c].CanWrite || !properties[c].CanRead) { continue; }
                properties[c].SetValue(buffer, DeSerialize(dataStream, properties[c].PropertyType), null);
            }

            //clean up
            dataStream.Dispose();

            //return the instance created from the type specified
            return buffer;
        }
        #endregion

        #region Enum
        public static void SerializeEnum(object value, Stream stream) { 
            //serialize the raw integer value of the enum
            stream.WriteByte((byte)SerializeType.Enum);
            SerializeLength((ulong)(int)value, stream);
        }
        public static Enum DeSerializeEnum(Stream stream, Type enumType) { 
            //read the type
            SerializeType type = (SerializeType)stream.ReadByte();

            //valid?
            if (type != SerializeType.Enum) { return null; }

            //return the value
            return (Enum)Enum.Parse(enumType, DeSerializeLength(stream).ToString());
        }
        #endregion

        #region Numerics
        public static void SerializeNumeric<T>(T value, Stream stream)
        { 
            //define the type of the serialization object
            SerializeType type = getIntSerializerType(value);

            #region Numerical serialization
            //define the data to write
            MemoryStream data = new MemoryStream();

            //is the value a decimal?
            if (type == SerializeType.Decimal) {
                //split up the float into 2 numbers (one for integer and one for the decimal)
                string[] sp = Convert.ToDecimal(value).ToString().Split('.');
                if (sp.Length == 1) {
                    sp = new string[] { sp[0], "0" };
                }

                //write the numbers in two parts
                SerializeLength(Convert.ToUInt64(sp[0]), data);
                SerializeLength(Convert.ToUInt64(sp[1]), data);
            }
            else {
                //serialize as an integer
                SerializeLength(Convert.ToUInt64(value), data);
            }
            #endregion

            //read the raw bytes from the data for the object
            byte[] raw = new byte[data.Length];
            data.Position = 0;
            data.Read(raw, 0, raw.Length);

            //serialize
            SerializeData(type, raw, stream);
        }
        public static void DeSerializeNumeric(ref object value, Stream stream) {
            //get the type of the value to compare with the read data
            SerializeType type = getIntSerializerType(value);
            if (type == SerializeType.Null) { value = null; return; }
            
            //read the raw integer data
            SerializeType compareType;
            byte[] data = DeSerializeData(stream, out compareType);
            if (compareType != type) { value = null; return; }
            
            //convert the data to a string
            Stream dataStr = new MemoryStream();
            dataStr.Write(data, 0, data.Length);
            dataStr.Position = 0;

            //normal integer?
            if (type != SerializeType.Decimal) { 
                //read the value
                ulong dataInt = DeSerializeLength(dataStr);

                //convert to the appropriate type
                switch (type) {
                    case SerializeType.Int8: value = (sbyte)dataInt; return;
                    case SerializeType.UInt8: value = (byte)dataInt; return;
                    case SerializeType.Int16: value = (short)dataInt; return;
                    case SerializeType.UInt16: value = (ushort)dataInt; return;
                    case SerializeType.Int32: value = (int)dataInt; return;
                    case SerializeType.UInt32: value = (uint)dataInt; return;
                    case SerializeType.Int64: value = (long)dataInt; return;
                    case SerializeType.UInt64: value = (ulong)dataInt; return;
                }

                //invalid
                value = null;
                return;
            }

            //read the decimal 2 parts
            ulong integer = DeSerializeLength(dataStr);
            ulong decimalInt = DeSerializeLength(dataStr);

            //set the value to the decimal value
            value = Convert.ToDecimal(integer.ToString() + "." + decimalInt.ToString());
        }

        /*These are just clones of SerializeSByte*/
        public static void SerializeSByte(sbyte value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeByte(byte value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeInt16(short value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeUInt16(ushort value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeInt32(int value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeUInt32(uint value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeInt64(long value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeUInt64(ulong value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeDouble(double value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeSingle(float value, Stream stream) {
            SerializeNumeric(value, stream);
        }
        public static void SerializeDecimal(decimal value, Stream stream) {
            SerializeNumeric(value, stream);
        }

        /*These are just clones of DeSerializeSByte*/
        public static sbyte DeSerializeSByte(Stream stream) {
            object o = (sbyte)0;
            DeSerializeNumeric(ref o, stream);
            return (sbyte)o;
        }
        public static byte DeSerializeByte(Stream stream) {
            object o = (byte)0;
            DeSerializeNumeric(ref o, stream);
            return (byte)o;
        }
        public static short DeSerializeInt16(Stream stream) {
            object o = (short)0;
            DeSerializeNumeric(ref o, stream);
            return (short)o;
        }
        public static ushort DeSerializeUInt16(Stream stream) {
            object o = (ushort)0;
            DeSerializeNumeric(ref o, stream);
            return (ushort)o;
        }
        public static int DeSerializeInt32(Stream stream) {
            object o = (int)0;
            DeSerializeNumeric(ref o, stream);
            return (int)o;
        }
        public static uint DeSerializeUInt32(Stream stream) {
            object o = (uint)0;
            DeSerializeNumeric(ref o, stream);
            return (uint)o;
        }
        public static long DeSerializeInt64(Stream stream) {
            object o = (long)0;
            DeSerializeNumeric(ref o, stream);
            return (long)o;
        }
        public static ulong DeSerializeUInt64(Stream stream) {
            object o = (ulong)0;
            DeSerializeNumeric(ref o, stream);
            return (ulong)o;
        }
        public static double DeSerializeDouble(Stream stream) {
            object o = (double)0;
            DeSerializeNumeric(ref o, stream);
            return (double)(decimal)o;
        }
        public static float DeSerializeSingle(Stream stream) {
            object o = (float)0;
            DeSerializeNumeric(ref o, stream);
            return (float)(decimal)o;
        }
        public static decimal DeSerializeDecimal(Stream stream) {
            object o = (decimal)0;
            DeSerializeNumeric(ref o, stream);
            return (decimal)o;
        }

        private static SerializeType getIntSerializerType(object value) {
            //decimal?
            if (value is float ||
                value is decimal ||
                value is double) { return SerializeType.Decimal; }

            //get the serializer type associated with the integer
            else if (value is sbyte) { return SerializeType.Int8; }
            else if (value is byte) { return SerializeType.UInt8; }
            else if (value is short) { return SerializeType.Int16; }
            else if (value is ushort) { return SerializeType.UInt16; }
            else if (value is int) { return SerializeType.Int32; }
            else if (value is uint) { return SerializeType.UInt32; }
            else if (value is long) { return SerializeType.Int64; }
            else if (value is ulong) { return SerializeType.UInt64; }

            //not detected
            return SerializeType.Null;
        }
        private static object createNumeric(SerializeType type) {
            switch (type) {
                case SerializeType.Int8: return (sbyte)0;
                case SerializeType.Int16: return (short)0;
                case SerializeType.Int32: return (int)0;
                case SerializeType.Int64: return (long)0;
                case SerializeType.UInt8: return (byte)0;
                case SerializeType.UInt16: return (ushort)0;
                case SerializeType.UInt32: return (uint)0;
                case SerializeType.UInt64: return (ulong)0;
            }

            //not found
            return null;
        }
        #endregion
    }

    /*
     * Functions from personal project but reduced for the requirements of this 
     * code file
    */
    public static class Helpers {
        public static bool CompareBytes(byte[] x, byte[] y) {
            //check length
            if (x.Length != y.Length) { return false; }

            //check for inconsistant values
            for (int c = 0; c < x.Length; c++) {
                if (x[c] != y[c]) { return false; }
            }

            //they're equal
            return true;
        }
        public static string ChangeNumericBase(int fromBase, int destBase, string source) {
            //this function has been completely re-written just for support for base 10 - 2 conversion only
                        
            //only conversion from base 10 and 2 are supported
            if (fromBase != 10 && fromBase != 2 &&
               destBase != 10 && destBase != 2) {
                   return null;
            }

            //same base?
            if (fromBase == destBase) { return source; }

            //base10-2 conversion
            if (fromBase == 10) {
                return Convert.ToString(Convert.ToInt64(source), 2);
            }
            else { 
                //binary to base10 conversion
                return Convert.ToUInt64(source, 2).ToString();
            }

        }
    }
    
    public interface ISerializer {
        void Serialize(object value, Stream stream, bool deep);

        object DeSerialize(Stream stream, Type objType);

        bool Accepts(object value, bool deep);

        byte TypeToken { get; }
        string Name { get; }
    }
    public enum SerializeType : byte {
        Null = 0,

        //integers (uses 1-19)
        Int8 = 1,
        UInt8 = 2,
        Int16 = 3,
        UInt16 = 4,
        Int32 = 5,
        UInt32 = 6,
        Int64 = 7,
        UInt64 = 8,
        Decimal = 11,

        Boolean = 21,
        Char = 22,
        String = 23,

        Object = 31,
        DateTime = 32,
        Array = 33,
        Enum = 34
    }
}